<?php
namespace PrestaShop\Module\LCVPrestaConnector;

use Exception;
use LCVPrestaConnector;
use PrestaShop\Module\LCVPrestaConnector\SyncProgress;

/**
 * Configuration du module
 */
class Syncer
{
    /**
     * Module
     */
    public LCVPrestaConnector $module;

    /**
     * Indique si nous sommes en mode CRON
     */
    private $isCron = false;

    /**
     * Pont vers le backoffice
     */
    public BridgeAdapter $bridge;

    /**
     * Lien vers le fichier d'audit
     */
    private mixed $auditFileHandler = null;

    /**
     * Cache de références fournisseur
     * null quand pas encore chargé
     * 
     * @var SupplierRefManager
     */
    private ?SupplierRefManager $refMngr = null;

    /**
     * Instanciation
     * 
     * @param LCVPrestaConnector $module Module
     * @param bool $isCron Indique si nous sommes en mode CRON
     */
    public function __construct(LCVPrestaConnector $module, $isCron = false)
    {
        // Déclenche la mise à jour de la base si besoin
        $module->dbUpgrade();

        $this->module = $module;        
        $this->bridge = $module->getBridge();
        $this->isCron = $isCron;
    }

    /**
     * Obtient le nom du verrou
     * 
     */
    public function getLockName() : string
    {
        return 'syncer-' . $this->module->name . '-' . str_replace('.', '_', $_SERVER['HTTP_HOST']);
    }

    /**
     * Obtient le nom du verrou
     * 
     */
    public static function constructLockName(\LCVPrestaConnector $module) : string
    {
        return 'syncer-' . $module->name . '-' . str_replace('.', '_', $_SERVER['HTTP_HOST']);
    }

    /**
     * Essaye de prendre le verrou de traitement ou lève une exception
     * 
     * @param bool $throwException Lève une exception si le verrou est déjà pris
     * @return bool Vrai si le verrou est pris, faux sinon
     */
    private function lock(bool $throwException = true) : bool
    {
        $db = \Db::getInstance();

        $result = $db->getValue('SELECT GET_LOCK(\'' . $this->getLockName() . '\', 30)');
        if ($result === false)
            throw new \Exception("Impossible de prendre le verrou de synchronisation : ".$db->getMsgError());
        
        // On a pris le verrou ?
        if (!$result)
        {
            // le verrou est déjà pris, on ne fait rien
            if ($throwException)                
                throw new \Exception("Une synchronisation est déjà en cours !");
            return false;
        }

        return true;    // verrou pris !
    }

    /**
     * Déverrouille le traitement
     */
    private function unlock()
    {
        $db = \Db::getInstance();
        // on libère le verrou
        $lockName = $this->getLockName();
        // et si on y arrive pas, on le mentionne dans le journal de Prestashop
        if (!$db->getValue('SELECT RELEASE_LOCK(\'' . $lockName . '\')'))
            \PrestaShopLogger::addLog('Impossible de libérer le verrou ' . $lockName . ' : ' . $db->getMsgError(), 3);
    }


    /**
     * Obtient le gestionnaire de références fournisseur
     * 
     * @return SupplierRefManager
     */
    public function getReferencesManager() : SupplierRefManager
    {
        if (is_null($this->refMngr))
            $this->refMngr = new SupplierRefManager($this->module->getCfg()->get(SyncConfiguration::CFG_SUPPLIER));
        return $this->refMngr;
    }

    /**
     * Resynchronise un seul produit particulier
     * 
     * @param int $id_product Identifiant du produit
     */
    public function syncProduct($id_product)
    {
        $cfg = $this->module->getCfg();

        // On précharge certaines données de configuration que nous allons qui nous seront utiles par la suite
        $cfg->preload([
            SyncConfiguration::CFG_SUPPLIER,
            SyncConfiguration::CFG_ACTIVE,
            SyncConfiguration::CFG_WORKMODE,
        ]);

        // Si le module n'est pas actif : pas de synchro !
        if (!$cfg->get(SyncConfiguration::CFG_ACTIVE))
            throw new \Exception('La synchronisation désactivée.');
 
        // on va passer par un verrou pour éviter de lancer plusieurs fois le traitement        
        if (!$this->lock())
        {
            // le verrou est déjà pris, on ne fait rien
            throw new \Exception("Une syncronisation est déjà en cours !");
        }

        try
        {
            $sync_progress = new SyncProgress($cfg, 1);
            $sync_progress->nextStep('Synchronisation du produit #' . $id_product);

            // On va chercher les références fournisseurs
            $refMngr = new SupplierRefManager($cfg->get(SyncConfiguration::CFG_SUPPLIER), [$id_product]);
            $productRef = $refMngr->translateProductId($id_product);
            if (!$productRef)
                throw new \Exception("Produit non géré par le backoffice " . $this->bridge->getId());
            
            $syncerProducts = new SyncerProducts($this);            
            $syncerProducts->syncProducts(
                $this->module->getBridge()->pullProducts($productRef)
            );
        }
        finally
        {
            // on libère le verrou
            $this->unlock();
        }
    }

    /**
     * Exporte une commande particulière
     * 
     * @param string $order_reference Référence de la commande
     */
    public function exportOrder(string $order_reference)
    {
        $cfg = $this->module->getCfg();

        // On précharge certaines données de configuration que nous allons qui nous seront utiles par la suite
        $cfg->preload([
            SyncConfiguration::CFG_SUPPLIER,
            SyncConfiguration::CFG_ACTIVE,
            SyncConfiguration::CFG_VOUCHER_IMPORT_TYPES,
            SyncConfiguration::CFG_EXPORT_CUSTOMERS,            
        ]);

        // Si le module n'est pas actif : pas de synchro !
        if (!$cfg->get(SyncConfiguration::CFG_ACTIVE))
            throw new \Exception('La synchronisation désactivée.');
 
        // on va passer par un verrou pour éviter de lancer plusieurs fois le traitement
        if (!$this->lock())
        {
            // le verrou est déjà pris, on ne fait rien
            throw new \Exception("Une syncronisation est déjà en cours !");
        }

        try
        {
            $sync_progress = new SyncProgress($cfg, 1);
            $sync_progress->nextStep('Synchronisation de la commande #' . $order_reference);

            $syncerOrders = new SyncerOrders($this); 
            $syncerOrders->exportOrder(
                $order_reference
            );
        }
        finally
        {
            // on libère le verrou
            $this->unlock();
        }
    }
    
    /**
     * Obtient la progression de la synchronisation
     */
    public static function getSyncProgress(\LCVPrestaConnector $module) : SyncProgress|null
    {
        // On vérifie si le verrou est positionné
        $db = \Db::getInstance();
        
        $result = $db->getValue('SELECT IS_USED_LOCK(\'' . self::constructLockName($module) . '\')');
        if ($result === false)
            throw new \Exception("Impossible de vérifier le verrou de synchronisation : ".$db->getMsgError());
        
        // On a pris le verrou ?
        if ($result)
        {
            // oui, on renvoi la progression qui est stockée dans la configuration
            return SyncProgress::fromArray($module->getCfg()->get(SyncConfiguration::CFG_SYNC_STATE));
        }
        else    // Non = pas de synchro
            return null;
    }

    /**
     * Synchronise les données
     */
    public function sync()
    {
        try
        {
            // On va passer par un verrou pour éviter de lancer plusieurs fois le traitement
            if (!$this->lock(false)) {
                // le verrou est déjà pris, on ne fait rien
                echo "Traitement déjà en cours !";
                return;
            }

            // Configuration
            $cfg = $this->module->getCfg();

            // Mise à jour de l'état
            $sync_progress = new SyncProgress($cfg, 6);
            
            // Etape préliminaire : tâches quotidiennes !
            // Toutes les 24h, on regarde s'il n'y a pas une mise à jour du module
            $lastCheck = \Configuration::get($this->module->name . '_last_update_check');
            if (!$lastCheck || (time() - $lastCheck) > 86400)
            {                
                $sync_progress->addStep('Vérification des mises à jour');
                $cfg->set(SyncConfiguration::CFG_SYNC_STATE, $sync_progress->toArray());

                // On enregistre la date de la dernière vérification
                \Configuration::updateValue($this->module->name . '_last_update_check', time());

                $updater = new AutoUpdater($this->module);
                $updater->searchUpdate();
            }

            // Si le module n'est pas actif : pas de synchro !
            if (!$cfg->get(SyncConfiguration::CFG_ACTIVE))
            {
                $this->audit('[WRN] La synchronisation désactivée.');
                return;
            }

            // preload cfg
            $cfg->preload([
                SyncConfiguration::CFG_ACTIVE,               
                SyncConfiguration::CFG_WORKMODE,
                SyncConfiguration::CFG_SUPPLIER, 
                SyncConfiguration::CFG_SYNC_STATE,
                SyncConfiguration::CFG_SYNC_POSITIONS,
                SyncConfiguration::CFG_VOUCHER_IMPORT_TYPES,
                SyncConfiguration::CFG_IMPORT_CUSTOMERS,
                SyncConfiguration::CFG_EXPORT_CUSTOMERS,
            ]);
            
            // Ensuite, on regarde si dans la configuration, nous avons l'identifiant du fournisseur pour ce module
            $supplierId = $cfg->get(SyncConfiguration::CFG_SUPPLIER);
            if ($supplierId)
            {
                // Chargement du fournisseur
                $supplier = new \Supplier($supplierId);
                if ($supplier->id != $supplierId)
                {
                    // On log dans prestashop
                    \PrestaShopLogger::addLog('Impossible de charger le fournisseur de référence #' . $supplierId . ' => recréation ! ' . $supplier->name, 3, null, "supplier", $supplierId);
                    $supplierId = 0;
                }
            }
            if (!$supplierId)
            {
                // Autrement, on le crée automatiquement
                $supplier = new \Supplier();
                // Le nom du fournisseur est le nom qui précède "LCVPrestaConnector" dans le nom du module, avec une majuscule                
                $supplier->name = ucfirst(substr($this->module->name, 0, strpos($this->module->name, 'LCVPrestaConnector')));
                if ($supplier->name == '')
                    $supplier->name = 'DEV';
                $supplier->add();
                if (!$supplier->id)
                    throw new Exception('Impossible de créer le fournisseur de référence !');
                $cfg->set(SyncConfiguration::CFG_SUPPLIER, $supplier->id);
                // On journalise la création du fournisseur dans les journaux de Prestashop
                \PrestaShopLogger::addLog('Création du fournisseur ' . $supplier->name, 1, null, "supplier", $supplier->id);
            }
            
            // Etape 1 : les clients ?
            $sync_progress->nextStep('Clients');
            if (
                $cfg->get(SyncConfiguration::CFG_IMPORT_CUSTOMERS) ||
                $cfg->get(SyncConfiguration::CFG_EXPORT_CUSTOMERS))
            {
                $syncerCustomers = new SyncerCustomers($this);
                $syncerCustomers->sync();
            }
            
            // Etape 2 : on transmet les commandes
            $sync_progress->nextStep('Commandes');
            $syncerOrders = new SyncerOrders($this);
            $syncerOrders->sync();
            
            // Etape 3 : on aspire le catalogue si workmode 1
            $sync_progress->nextStep('Catalogue');
            if ($cfg->get(SyncConfiguration::CFG_WORKMODE) == 1)
            {
                $syncerProducts = new SyncerProducts($this);
                $syncerProducts->sync();
            }

            // Etape 4 : on aspire les stocks et les prix
            $sync_progress->nextStep('Stocks & prix');
            $syncerStocks = new SyncerStocks($this);
            $syncerStocks->sync();

            // Etape 5 : les promotions ?
            $sync_progress->nextStep('Promotions');
            $syncerDiscounts = new SyncerDiscounts($this);
            $syncerDiscounts->sync();            

            // Etape 6 : les bons de réduction ?
            // Seulement si on importe les bons cadeaux ou les avoirs
            $sync_progress->nextStep('Bons de réduction');
            $voucherTypes = [];
            foreach ($cfg->get(SyncConfiguration::CFG_VOUCHER_IMPORT_TYPES) ?? [] as $type => $actif)
                if ($actif)
                    $voucherTypes[$type] = true;
            if (count($voucherTypes) > 0)
            {
                $syncerVouchers = new SyncerVouchers($this);
                $syncerVouchers->sync();
            }

            // Fin !
            $this->audit('[INFO] Synchronisation terminée');
        }
        catch (Exception $e)
        {
            // Journalisation de l'erreur
            \PrestaShopLogger::addLog('Erreur à la synchronisation : '.$e->getMessage(), 3, null, $this->module->name); 
            $this->audit('[FATAL] ' . $e->getMessage());
        }
        finally
        {
            // on libère le verrou
            $this->unlock();            
        }
    }

    /**
     * Audit des actions
     * 
     * @param string $message Message à journaliser     
     */
    public function audit(string $message)
    {
        if (!$message)
            return; // rien à faire 

        // Journalisation de l'audit, séparé, dans un fichier à part
        // le fichier est il ouvert ?
        if (!$this->auditFileHandler)
        {
            // sinon, on l'ouvre en mode ajout dans la racine /var/audits
            // si le répertoire n'existe pas, on le crée            
            $path = _PS_ROOT_DIR_ . '/var/audits/'.$this->module->name;
            if (!is_dir($path))
                mkdir($path, 0777, true);

            // On supprime les fichiers d'audit qui ont plus de 60 jours
            $files = scandir($path);
            foreach ($files as $file)
            {
                if (is_file($path . '/' . $file) && (time() - filemtime($path . '/' . $file)) > 60 * 86400)
                    unlink($path . '/' . $file);
            }

            // Le fichier est nommé audit-AAAAMMJJHHmmSS.log
            $this->auditFileHandler = fopen($path . '/audit-' . date('Ymd') . '.log', 'a');
            if (!$this->auditFileHandler)
                throw new Exception('Impossible d\'ouvrir le fichier d\'audit');            
        
            fwrite($this->auditFileHandler, date('Y-m-d H:i:s') . ' - Nouvelle synchronisation'."\n");
        }

        // On écrit le message dans le fichier
        fwrite($this->auditFileHandler, date('Y-m-d H:i:s') . ' - ' . $message . "\n");

        // Si nous sommes en mode DEBUG, alors on sort le message sur la sortie standard si nous sommes un tâche CRON
        if (_PS_MODE_DEV_ && $this->isCron)
            echo $message . "\n"; 
    }    
}